


.include "at90s2313.inc"

; input impuls counter
.def	imp_in	= r10
; output impuls counter
.def	imp_out	= r11
; # of imps in a row
.def	imp_cnt	= r12
; minimum in imps
.def	imp_high	= r13
; maximum in imps
.def	imp_max	= r14

; reset alarm delay and status blink delay
.def	rcnt	= r9

; temp registers for can rx interrupts
.def	ccmd	= r21
.def	cpar	= r22

; temp registers for timer interrupt
.def	tmp	= r24
.def	tt_tmp	= r23

; trigger flags
.def	trx	= r25
.equ	tr0_flg	= 0
.equ	tr1_flg	= 1
.equ	tr0_flg_cur	= 2
.equ	tr1_flg_cur	= 3

; SREG push/pop register
.def	rex	= r15

; MCP pins
.equ	SCK	= PB7
.equ	MISO	= PB6
.equ	MOSI	= PB5
.equ	CANCS	= PB4
.equ	RX0	= PD2
.equ	RX1	= PD3
.equ	TX0	= PD4
.equ	TX1	= PD5
.equ	TX2	= PD6

; can extended identifier
.equ	aext	= 0b10101000
; can broadcast address
.equ	aall	= 0b10101010

; can commando codes
.equ	cmd_prp	= 0x01
.equ	cmd_tr0	= 0x80
.equ	cmd_tr1	= 0x81
.equ	cmd_tr0s	= 0x90
.equ	cmd_tr1s	= 0x91


; status led
.macro	led_on
	cbi	PORTB,	PB2
.endmacro

.macro	led_off
	sbi	PORTB,	PB2
.endmacro


.cseg

.org	$000
	rjmp	start

.org	INT0addr
	rjmp	can_rx0

.org	INT1addr
	rjmp	can_rx1

.org	TCP1addr
	reti

.org	TOF0addr
	rjmp	timer

.org	ACPaddr


; *
; *  analog comparator ( counts incoming transitions )

lsr_int:
	in	rex,	SREG

	inc	imp_in

	out	SREG,	rex
	reti


; *
; *  timer 0  ( toggle the laser beam on/off & check alarm )

timer:
	in	rex,	SREG

	; if laser is "armed"
	sbrs	trx,	tr1_flg
	rjmp	timer_ntgl

	; toggle LED on / off
	in	tmp,	TCCR1A
	ldi	tt_tmp,	( 1 << COM1A1 ) | ( 1 << COM1A0 )
	eor	tmp,	tt_tmp
	out	TCCR1A,	tmp
timer_ntgl:

	; count out pulse
	inc	imp_out
	cp	imp_out,	imp_cnt
	brcs	timer_fin

	; compare in pulse
	cp	imp_in,	imp_high
	brcs	lsr_alarm
	cp	imp_max,	imp_in
	brcs	lsr_alarm

	led_off

	; reset delay
	inc	rcnt
	sbrs	rcnt,	6
	rjmp	timer_clr
	clr	rcnt

	; alarm off
	rcall	tr0_set0

	; status blink
	led_on

	rjmp	timer_clr

lsr_alarm:
	; laser alarm
	led_on
	rcall	tr0_set1
	clr	rcnt

timer_clr:
	; reset puls counters
	clr	imp_in
	clr	imp_out

timer_fin:
	out	SREG,	rex
	reti


; *
; *  can rx interrupts

can_rx1:
	in	rex,	SREG
	push	r16
	; set load rx buffer 0 command
	ldi	r16,	0b10010100
	rjmp	can_rx
can_rx0:
	in	rex,	SREG
	push	r16
	; set load rx buffer 1 command
	ldi	r16,	0b10010000
can_rx:
	; load rx buffer
	cbi	PORTB,	CANCS
	rcall	spi_write
	; skip address bytes ( hardware filter activated )
	rcall	spi_read
	rcall	spi_read
	; read command
	rcall	spi_read
	mov	ccmd,	r16
	; read parameter
	rcall	spi_read
	mov	cpar,	r16
	; read remote frame request flag
	rcall	spi_read
	bst	r16,	6
	sbi	PORTB,	CANCS

can_rx_prp:
	; propagate station id ?
	cpi	ccmd,	cmd_prp
	brne	can_rx_tr0
	brtc	can_rx_fin
	cbi	PORTD,	TX0
	sbi	PORTD,	TX0
	rjmp	can_rx_fin
can_rx_tr0:
	; request trigger 0 ?
	cpi	ccmd,	cmd_tr0
	brne	can_rx_tr1
	brtc	can_rx_fin
	rcall	tr0_snd
	rjmp	can_rx_fin
can_rx_tr1:
	; request trigger 1 ?
	cpi	ccmd,	cmd_tr1
	brne	can_rx_tr0s
	brtc	can_rx_fin
	rcall	tr1_snd
	rjmp	can_rx_fin
can_rx_tr0s:
	; set trigger 0 ?
	cpi	ccmd,	cmd_tr0s
	brne	can_rx_tr1s
	brts	can_rx_fin
	sbrs	trx,	tr0_flg
	rjmp	can_rx_fin
	cbr	trx,	( 1 << tr0_flg )
	rcall	tr0_snd
	rjmp	can_rx_fin
can_rx_tr1s:
	; set trigger 1 ?
	cpi	ccmd,	cmd_tr1s
	brne	can_rx_fin
	brts	can_rx_fin
	bst	cpar,	0
	rcall	tr1_set
can_rx_fin:
	pop	r16
	out	SREG,	rex
	reti


; *
; *  program entry point

start:
	; init stack pointer
	ldi	r16,	RAMEND
	out	SPL,	r16

	; setup ports
	ldi	r16,	0b01110110
	out	PORTB,	r16
	ldi	r16,	0b10111100
	out	DDRB,	r16
	ldi	r16,	0b01110000
	out	PORTD,	r16
	ldi	r16,	0b01110000
	out	DDRD,	r16

	; reset triggers
	clr	trx

	; reset delay counter
	clr	rcnt

	; setup impuls counters
	clr	imp_in
	clr	imp_out
	ldi	ZL,	laser_cnts
	rcall	ee_read
	mov	imp_cnt,	r16
	rcall	ee_read
	mov	imp_high,	r16
	rcall	ee_read
	mov	imp_max,	r16

	; enable timer
	rcall	ee_read
	out	TCCR0,	r16
	ldi	r16,	( 1 << TOIE0 )
	out	TIMSK,	r16

	; enable sleep mode ( idle )
	ldi	r16,	( 1 << SE )
	out	MCUCR,	r16

	; int0/1 zero level triggered
	ldi	r16,	( 1 << INT0 ) | ( 1 << INT1 )
	out	GIMSK,	r16

	; enable analog comparator interrupt
	ldi	r16,	( 1 << ACIE )
	out	ACSR,	r16

	; setup mcp2515
	ldi	ZL,	can_init
start_can:
	rcall	ee_read
	tst	r16
	breq	start_can_done

	cbi	PORTB,	CANCS
	rcall	spi_write

	rcall	ee_read
	tst	r16
	breq	start_can_finish_command

	mov	r1,	r16
start_can_write_data:
	rcall	ee_read
	rcall	spi_write
	dec	r1
	brne	start_can_write_data
start_can_finish_command:
	sbi	PORTB,	CANCS
	rjmp	start_can
start_can_done:

	; propagate station id
	cbi	PORTD,	TX0
	sbi	PORTD,	TX0

	; start laser
	rcall	tr1_set1

	; enable interrupts
	sei

loop:
	; wait for something to happen
	sleep
	rjmp	loop


; *
; *  trigger 0  ( alarm flag )

tr0_set:
	brtc	tr0_set0
tr0_set1:
	sbrc	trx,	tr0_flg_cur
	ret

	sbr	trx,	( 1 << tr0_flg )
	rcall	tr0_snd
	sbr	trx,	( 1 << tr0_flg_cur )
	ret

tr0_set0:
	sbrs	trx,	tr0_flg_cur
	ret

	cbr	trx,	( 1 << tr0_flg )
	rcall	tr0_snd
	cbr	trx,	( 1 << tr0_flg_cur )
	ret

tr0_snd:
	push	r16

	; update can message buffer with current trigger flag
	cbi	PORTB,	CANCS
	ldi	r16,	0b00000010
	rcall	spi_write
	ldi	r16,	0b01000100
	rcall	spi_write
	clr	r16
	bst	trx,	tr0_flg
	bld	r16,	0
	bst	trx,	tr0_flg_cur
	bld	r16,	1
	rcall	spi_write
	sbi	PORTB,	CANCS

	; request to send
	cbi	PORTD,	TX1
	sbi	PORTD,	TX1

	pop	r16
	ret


; *
; *  trigger 1  ( laser beam activation )

tr1_set:
	brtc	tr1_set0
tr1_set1:
	sbrc	trx,	tr1_flg_cur
	ret

	sbr	trx,	( 1 << tr1_flg )
	rcall	tr1_snd
	sbr	trx,	( 1 << tr1_flg_cur )

	; activate laser beam modulation
	push	r16
	push	ZL
	ldi	ZL,	laser_freq
	ldi	r16,	( 1 << COM1A0 )
	out	TCCR1A,	r16
	rcall	ee_read
	ori	r16,	( 1 << CTC1 )
	out	TCCR1B,	r16
	rcall	ee_read
	out	OCR1AH,	r16
	rcall	ee_read
	out	OCR1AL,	r16
	clr	r16
	out	TCNT1H,	r16
	out	TCNT1L,	r16
	pop	ZL
	pop	r16

	ret
tr1_set0:
	sbrs	trx,	tr1_flg_cur
	ret

	; stop beamer
	push	r16
	ldi	r16,	( 1 << COM1A1 )
	out	TCCR1A,	r16
	pop	r16

	cbr	trx,	( 1 << tr1_flg )
	rcall	tr1_snd
	cbr	trx,	( 1 << tr1_flg_cur )
	ret

tr1_snd:
	push	r16
	; update can message buffer
	cbi	PORTB,	CANCS
	ldi	r16,	0b00000010
	rcall	spi_write
	ldi	r16,	0b01010100
	rcall	spi_write
	clr	r16
	bst	trx,	tr1_flg
	bld	r16,	0
	bst	trx,	tr1_flg_cur
	bld	r16,	1
	rcall	spi_write
	sbi	PORTB,	CANCS
	; request to send
	cbi	PORTD,	TX2
	sbi	PORTD,	TX2
	pop	r16
	ret



; *
; *  software spi

spi_write:
	push	r17
	ldi	r17,	8
spi_write_1:
	cbi	PORTB,	SCK
	rol	r16
	cbi	PORTB,	MOSI
	brcc	spi_write_2
	sbi	PORTB,	MOSI
spi_write_2:
	sbi	PORTB,	SCK
	dec	r17
	brne	spi_write_1
	cbi	PORTB,	SCK
	pop	r17
	ret

spi_read:
	push	r17
	ldi	r17,	8
spi_read_1:
	sbi	PORTB,	SCK
	clc
	sbic	PINB,	MISO
	sec
	rol	r16
	cbi	PORTB,	SCK
	dec	r17
	brne	spi_read_1
	pop	r17
	ret


; *
; *  eeprom

ee_read:
	out	EEAR,	ZL
	sbi	EECR,	EERE
	in	r16,	EEDR
	inc	ZL
	ret


.eseg

laser_freq:
; laser frequency : 4000000 / 1 / 55 = 36000
.db	0x01, 0x00, 55

laser_cnts:
; toggle beam 31 times
; alarm if imp_in < 5 or imp_in > 40
; toggle frequency : 4000000 / 64 / 256 = 244
; this means: response time ~ 1/5 seconds and minimum sensitivity
.db	31, 5, 40, 0b011


; *
; *  setup sequence for mcp2515
; *
; *  command, datalength, data, data, ...

can_init:
; enter configuration mode
.db	0b00000010, 2, 0x0f, 0x80
; set pin functs
.db	0b00000010, 3, 0x0c, 0b00001111,0b00000111
; enable receive filter
.db	0b00000010, 2, 0x60, 0b01001000
.db	0b00000010, 2, 0x70, 0b01001000
.db	0b00000010,13, 0x00, addr,aext,cmd_prp,0x00, aall,aext,cmd_prp,0x00, addr,aext,cmd_tr0,0x00
.db	0b00000010,13, 0x10, addr,aext,cmd_tr1,0x00, addr,aext,cmd_tr0s,0x00, addr,aext,cmd_tr1s,0x00
.db	0b00000010, 9, 0x20, 0xff,0xff,0xff,0x00, 0xff,0xff,0xff,0x00
; select baud rate
.db	0b00000010, 4, 0x28, 0b00000101,0b11110011,0b11000010
; fill transmit buffers
.db	0b00000010,15, 0x30, 0x01, addr, aext, cmd_prp, 0x20, 8, "laser ", lida, lidb
.db	0b00000010,15, 0x40, 0x00, addr, aext, cmd_tr0, 0x00, 8, "!alarm  "
.db	0b00000010,15, 0x50, 0x00, addr, aext, cmd_tr1, 0x00, 8, "active  "
; start normal operation
.db	0b00000010, 2, 0x0f, 0x00
; done
.db	0
